/* util.c - do statistical tests
 *
 * Copyright (C) 2001  Jochen Voss.  */

static const  char  rcsid[] = "$Id: util.c 5876 2004-08-14 10:36:54Z voss $";

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <limits.h>
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include <unistd.h>
extern char *optarg;
extern int  optind;
#endif
#if HAVE_ERRNO_H
#include <errno.h>
#else
extern  int  errno;
#endif
#include <assert.h>

#include "util.h"


const char *my_name;
static  double  alpha;


void
init_util (int argc, char **argv)
{
  const char *alpha_str;
  int  error_flag = 0;
  int  version_flag = 0;
  
  my_name = strrchr (argv[0], '/');
  if (my_name) {
    my_name += 1;
  } else {
    my_name = argv[0];
  }
  if (strncmp (my_name, "lt-", 3) == 0)  my_name += 3;
  my_name = strdup (my_name);

  alpha = 0.0001;
  alpha_str = getenv ("ALPHA");
  if (alpha_str) {
    char *tail;
    double  val;
	
    errno = 0;
    val = strtod (alpha_str, &tail);
    if ((errno == ERANGE || val <= 0 || val > 0.5) && *tail == '\0') {
      fprintf (stderr, "warning: invalid number ALPHA=\"%s\" in environment\n",
	       alpha_str);
    } else if (errno || tail == alpha_str || *tail != '\0') {
      fprintf (stderr,
	       "warning: malformed number ALPHA=\"%s\" in environment\n",
	       alpha_str);
    } else {
      alpha = val;
    }
  }

  do {
    int  c = getopt (argc, argv, "a:V");
    if (c == -1)  break;

    switch (c) {
    case 'a':
      {
	char *tail;
	double  val;
	
	errno = 0;
	val = strtod (optarg, &tail);
	if ((errno == ERANGE || val <= 0 || val > 0.5) && *tail == '\0') {
	  fprintf (stderr, "invalid number \"%s\"\n", optarg);
	  error_flag = 1;
	} else if (errno || tail == optarg || *tail != '\0') {
	  fprintf (stderr, "malformed number \"%s\"\n", optarg);
	  error_flag = 1;
	} else {
	  alpha = val;
	}
      }
      break;
    case 'V':
      version_flag = 1;
      break;
    default:
      error_flag = 1;
    }
  } while (! error_flag);

  if (argc != optind) {
    fputs ("too many arguments\n", stderr);
    error_flag = 1;
  }

  if (version_flag) {
    puts ("Jvrand " VERSION);
    puts ("Copyright 2001 Jochen Voss");
    puts ("\
Jvrand comes with NO WARRANTY, to the extent permitted by law.");
    puts ("\
You may redistribute copies of Jvrand under the terms of the GNU\n\
General Public License.  For more information about these matters,\n\
see the file named COPYING from the jvrand distribution.");
    if (! error_flag)  exit (0);
  }
  if (error_flag) {
    fprintf (stderr, "usage: %s [-V] [-a alpha]\n", my_name);
    exit (1);
  }
}

static double
Phi (double x)
/* The distribution function of the standard normal distribution.  */
{
  return  (1+erf (x/M_SQRT2))/2;
}

static double
limit (double q)
{
  double  l, r;
  
  assert (q < 0.5);

  l = -1;  r = 0;
  while (Phi(l) > q)  r = l, l -= 1;
  /* Phi(l) <= q && q < Phi(r) */
  while (l+1e-6 < r) {
    double  m = 0.5*(l+r);
    if (Phi(m) <= q) {
      l = m;
    } else {
      r = m;
    }
  }
  return  -l;
}

static unsigned
sample_size (double t, double v, double d)
/* Return the number of iterations needed to make effects of size D visible.
 * T is the 1-alpha/2 quantile, V the variance.  */
{
  double  x;

  x = t*sqrt(v)/d;
  assert (x*x <= UINT_MAX);
  return  x*x + 0.5;
}

void
binomial_test (b_test_fn f, void *client_data, double p, double dp,
	       const char *msg, int *fail_flag)
/* Test for the parameter P of a binomial distribution.  */
{
  double  e, s, x, t;
  unsigned  n, k, steps;

  printf ("%s: %s ... ", my_name, msg);
  fflush (NULL);
  
  assert (dp > 0);
  assert (0 <= p-dp && p+dp <= 1);
  t = limit (alpha/2);
  steps = sample_size (t, p*(1-p), dp);
  for (n=0, k=0; n<steps; ++n) {
    if (f (client_data))  ++k;
  }

  assert (n*p*(1-p) >= 9);
  e = n*p;
  s = sqrt (n*p*(1-p));
  x = (k-e)/s;
  
  if (fabs (x) <= t) {
    puts ("passed");
  } else {
    puts ("FAILED");
    fprintf (stderr,
	     "    Parameter %g is outside the confidence interval [%g; %g]\n"
	     "    This may be due to statistical fluctuations, so try again.\n"
	     "    In the long run this test should fail 1 of %d times.\n",
	     p, (k-t*s)/n, (k+t*s)/n, (int)(1/alpha+0.5));
    *fail_flag = 1;
  }
}

void
normal_test (n_test_fn f, void *client_data, double m, double dm, double v,
	     const char *msg, int *fail_flag)
/* Test for the mean M of a gaussion distribution with known variance V.  */
{
  double  s, t, d;
  unsigned  n, steps;

  printf ("%s: %s ... ", my_name, msg);
  fflush (NULL);

  assert (v > 0);
  t = limit (alpha/2);
  steps = sample_size (t, v, dm);
  s = 0;
  for (n=0; n<steps; ++n) {
    double  x = f (client_data);
    s += x-m;
  }
  s /= n;
  d = t*sqrt(v/n);
  
  if (fabs (s) <= d) {
    puts ("passed");
  } else {
    puts ("FAILED");
    fprintf (stderr,
	     "    Parameter %g is outside the confidence interval [%g; %g]\n"
	     "    This may be due to statistical fluctuations, so try again.\n"
	     "    In the long run this test should fail 1 of %d times.\n",
	     m, s-d, s+d, (int)(1/alpha+0.5));
    *fail_flag = 1;
  }
}
